/*
  adts

  Original work by Matteo Croce
  Small changes and fixes by Rafael Diniz

  Copyright (C) Matteo Croce and Rafael Diniz
 */
#include <string.h>
#include "adts.h"
#include "aacenc.h"

const int adts_sample_rates[] = {96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,7350,0,0,0};
const int id = 0; /* 0: MPEG-4, 1: MPEG-2 */
const int profile = 1;

/* buffer_fullness vars */
static int bit_reservoir_state = 0;
static unsigned short  mean_framelength = 0;
static unsigned short frl_divider = 0;
static unsigned short max_bit_reservoir = 0;

static int FindAdtsSRIndex(int sr)
{
    int i;
    for (i = 0; i < 16; i++) {
	if (sr == adts_sample_rates[i])
	    return i;
    }
    return 16 - 1;
}

void adts_hdr(char *adts, AACENC_CONFIG *config) {
    int srate_idx = FindAdtsSRIndex(config->sampleRate);
    
    /* SOURCE: http://blog.olivierlanglois.net/index.php/2008/09/12/aac_adts_header_buffer_fullness_field */
    unsigned short min_dec_in_size = (6144*config->nChannelsOut);
    mean_framelength = (unsigned short)((float) config->bitRate/config->sampleRate*1024);
    max_bit_reservoir = min_dec_in_size - mean_framelength;
    frl_divider = 32*config->nChannelsOut;
    
    memset(adts, 0, ADTS_HEADER_SIZE);
    
    adts[0] = 0xFF;	/* 8bits: syncword */
    adts[1] = 0xF0;	/* 4bits: syncword */
    adts[1] |= id << 3; /* 1bit:  mpeg id = 0 */
			/* 2bits: layer = 00 */
    adts[1] |= 1; 	/* 1bit:  protection absent (1 - means "no protection")*/
    adts[2] = ((profile << 6) & 0xC0);		/* 2bits: profile */
    adts[2] |= ((srate_idx << 2) & 0x3C);	/* 4b: sampling_frequency_index */
		/* 1b: private = 0 */
    adts[2] |= ((config->nChannelsOut >> 2) & 0x1); /* 1b: channel_configuration */
    adts[3] = ((config->nChannelsOut << 6) & 0xC0); /* 2b: channel_configuration */
}

void adts_hdr_up(char *adts, int size)
{
    unsigned short len = size + ADTS_HEADER_SIZE;

    bit_reservoir_state += mean_framelength - (size*8);
    if (bit_reservoir_state < 0) {
	bit_reservoir_state = 0;
    } else if (bit_reservoir_state > max_bit_reservoir) {
	bit_reservoir_state = max_bit_reservoir;
    }
    unsigned short buffer_fullness = bit_reservoir_state/frl_divider;
								    
    /* frame length, 13 bits */
    adts[3] &= 0xFC;
    adts[3] |= ((len >> 11) & 0x03);	/* 2b: aac_frame_length */
    adts[4] = len >> 3;			/* 8b: aac_frame_length */
    adts[5] = (len << 5) & 0xE0;	/* 3b: aac_frame_length */
    /* buffer fullness, 11 bits */
    adts[5] |= ((buffer_fullness >> 6) & 0x1F);	/* 5b: adts_buffer_fullness */
    adts[6] = (buffer_fullness << 2) & 0xFC;	/* 6b: adts_buffer_fullness */
						/* 2b: num_raw_data_blocks */
}
